/*	$Id: imuFilter.c,v 1.1 2007/04/16 14:11:06 sjoyeux Exp $ */


/**************************************************************
 **
 ** Copyright (c) 2004 LAAS/CNRS 
 **
 ** Thierry Peynot - 2004
 **
 ** $Source: /cvs/openrobots/modules/imu/codels/imuFilter.c,v $
 ** $Revision: 1.1 $
 ** $Date: 2007/04/16 14:11:06 $
 ** 
 **************************************************************/


/* 
 * Copyright (c) 2004 LAAS/CNRS                       --  Wed Apr 21 2004
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice,  this list of  conditions and the following disclaimer in
 *      the  documentation  and/or  other   materials provided  with  the
 *      distribution.
 *
 * THIS  SOFTWARE IS PROVIDED BY  THE  COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND  ANY  EXPRESS OR IMPLIED  WARRANTIES,  INCLUDING,  BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES  OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR  PURPOSE ARE DISCLAIMED. IN  NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR      CONTRIBUTORS  BE LIABLE FOR   ANY    DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF  SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE   OF THIS SOFTWARE, EVEN   IF ADVISED OF   THE POSSIBILITY OF SUCH
 * DAMAGE.
 */

#ifndef UNIX
#include <vxWorks.h>
#else
#include <portLib.h>
#endif
#include <taskLib.h>

#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <math.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <termios.h>

#include "commonStructLib.h"

#include "server/imuHeader.h"
#include "imuInt.h"

#include <stdlib.h>


/********************************************************************************
 *
 *  FILTER for the angular speeds and accelerations measured by the IMU
 *
********************************************************************************/




/*********** Global variables *******************************/

typedef double AngSpeeds[IMU_NUM_ANGSPEEDS];
typedef double Accel[IMU_NUM_ACCEL];

/* AngularSpeeds Filter Params */
static int widthFilter;
static float sigmaFilter;
static float*  weightsFilter = NULL;  /* weights array for "Gauss Filter" */
static float   weightsSumF;

/* Accelerations Filter Params */
static int widthFilter_accel;
static float sigmaFilter_accel;
static float*  weightsFilter_accel = NULL;  /* weights array for "Gauss Filter" */
static float   weightsSumF_accel;


/* AngularSpeeds buffer */
static AngSpeeds*  buffer = NULL;
static int k;			/* current index */
static int K;			/* sample number */

/* Accelerations buffer */
static Accel*  buffer_accel = NULL;
static int k_accel;			/* current index */
static int K_accel;			/* sample number */

/*************************************************************/



/* Local functions */
static float imuGauss (float x, float m, float sigma);




/**************************************************************************************
 *
 * Calculates the result of the Gauss function
 * Returns this result
 *
 **************************************************************************************/

static float
imuGauss (float    x,
	  float    m,        /* m = average (moyenne ou Esperance) */
	  float    sigma )    /* sigma^2 = variance  */

{
  return ( (1/sqrt(2.0*Pi*sigma*sigma)) * exp(-(x-m)*(x-m)/(2.0*sigma*sigma)) );
}





/**************************************************************************************
 *
 * Initializes static variables for Angular Speeds Filter
 *
 **************************************************************************************/

STATUS
imuFilterInit (int numSamplesUsedToFilterParam, /* width of Filter Gauss function */
	       float sigmaFilterParam)   /* sigma of Filter Gauss function */
{
  int i;


  widthFilter = numSamplesUsedToFilterParam;
  sigmaFilter = sigmaFilterParam;


  /* allocation of the angular speeds buffer */
  if (buffer != NULL) free(buffer);

  buffer = calloc ((widthFilter+1), sizeof (AngSpeeds));
  if ( buffer == NULL ) {
    return ERROR;
  }


  k = K = 0;

  /* PREPARATION of the parameters and structures for the filter ******************/
  
  /* allocation of the weights array */
  if (weightsFilter != NULL) free(weightsFilter);

  if ( (weightsFilter = calloc ((widthFilter+1), sizeof (float)) ) == NULL )
    return ERROR;
  
  /* Calculation of the "weights" determined by the Gaussian function */
  for (i = 0; i <= widthFilter; i++)
    weightsFilter[i] = imuGauss( (float) i, 0.0, sigmaFilter );
   /* the highest weight is weights[0]  */
  /***********************************************************************/
  weightsSumF = 0.0;


  return OK;
}




/**************************************************************************************
 *
 * Initializes static variables for Accelerations Filter
 *
 **************************************************************************************/

STATUS
imuFilterInit_accel (int numSamplesUsedToFilterParam_accel, /* width of Filter Gauss function */
		     float sigmaFilterParam_accel)   /* sigma of Filter Gauss function */
{
  int i;


  widthFilter_accel = numSamplesUsedToFilterParam_accel;
  sigmaFilter_accel = sigmaFilterParam_accel;


  /* allocation of the angular speeds buffer */
  if (buffer_accel != NULL) free(buffer_accel);

  buffer_accel = calloc ((widthFilter_accel+1), sizeof (Accel));
  if ( buffer_accel == NULL ) {
    return ERROR;
  }


  k_accel = K_accel = 0;

  /* PREPARATION of the parameters and structures for the filter ******************/
  
  /* allocation of the weights array */
  if (weightsFilter_accel != NULL) free(weightsFilter_accel);

  if ( (weightsFilter_accel = calloc ((widthFilter_accel+1), sizeof (float)) ) == NULL )
    return ERROR;
  
  /* Calculation of the "weights" determined by the Gaussian function */
  for (i = 0; i <= widthFilter_accel; i++)
    weightsFilter_accel[i] = imuGauss( (float) i, 0.0, sigmaFilter_accel );
   /* the highest weight is weights[0]  */
  /***********************************************************************/
  weightsSumF_accel = 0.0;


  return OK;
}








/**************************************************************************************
 *
 * Applies a "Gauss Average Filter" to the angular speeds signals given by IMU
 *
 * Input:  The angular speeds (or an array conataining that)
 * Output: array containing the angular speeds filtered (-> angSpeedsFiltered )
 *
 **************************************************************************************/


STATUS
imuGaussAverageFilter(double   omega_x,   /* Angular speeds given by IMU */
		      double   omega_y,
		      double   omega_z,
		      double   angSpeedsFiltered[IMU_NUM_ANGSPEEDS] ) /* output */

{
  int     i, j;
  /* i = number of the previous sample used to filter */  
  AngSpeeds samplesSum;


  /* Read the new sample values */
  buffer[k][IMU_OMEGA_X] = omega_x;
  buffer[k][IMU_OMEGA_Y] = omega_y;
  buffer[k][IMU_OMEGA_Z] = omega_z;



  /* *************** FILTER ************************************************/

  /* Weights are applied to the past samples (widthFilter before)
   * Then the current sample is replaced by the weighted average */
  
  /* init samplesSum array */
  for (j = 0; j < IMU_NUM_ANGSPEEDS; j++)
    samplesSum[j] = 0.0;

  /* update the sum of the weights */
  if ( K == 0 )  /* first sample */
    /* initial weightsSumF */
    weightsSumF = weightsFilter[0];
  else if (K <= widthFilter)
    /* the current sample is number K < widthFilter ********/
    weightsSumF += weightsFilter[K];

  /* compute the weighted sum of samples */
  if (K <= widthFilter)      /* k = K here */
  {
    for (i = 0; i <= K; i++)   /* i: before the current sample K to modify */
    {
      /* weighted sums */
      for (j = 0; j < IMU_NUM_ANGSPEEDS; j++)
	samplesSum[j] = samplesSum[j] + weightsFilter[i]*buffer[k-i][j];
    }
  }
  /* At this moment weightsSumF contains all the weights */
  else  /* the current sample is number K >= widthFilter  ***/
  { 
    /* all the weights can be applied every time */
    /* Thus, weightsSumF is always the same here */

    for (i = 0; i <= k; i++)   /* from k to beginning of buffer */
    {
      /* weighted sum */
      for (j = 0; j < IMU_NUM_ANGSPEEDS; j++)
	samplesSum[j] = samplesSum[j] + weightsFilter[i]*buffer[k-i][j];
    }
    for (i = k+1; i <= widthFilter; i++)   /* from end of buffer to k+1 */
    {
      for (j = 0; j < IMU_NUM_ANGSPEEDS; j++)
	samplesSum[j] = samplesSum[j] + 
	  weightsFilter[i]*buffer[widthFilter+1+k-i][j];
    }
  }

  /* the sample k is replaced by the weighted average */
  for (j = 0; j < IMU_NUM_ANGSPEEDS; j++)
    buffer[k][j] = samplesSum[j] / weightsSumF;

  /* *************** END OF FILTER ****************************************/

  
  /* Get the Filtered values -> Output */
  angSpeedsFiltered[IMU_OMEGA_X] = buffer[k][IMU_OMEGA_X];
  angSpeedsFiltered[IMU_OMEGA_Y] = buffer[k][IMU_OMEGA_Y];
  angSpeedsFiltered[IMU_OMEGA_Z] = buffer[k][IMU_OMEGA_Z];


  
  /* Change k: the number of the new sample in the buffer */
  k = (k+1)%(widthFilter+1);
  K++;


  return OK;
}






/**************************************************************************************
 *
 * Applies a "Gauss Average Filter" to the accelerations signals given by IMU
 *
 * Input:  The accelerations (or an array conataining that)
 * Output: array containing the accelerations filtered (-> accelFiltered )
 *
 **************************************************************************************/


STATUS
imuGaussAverageFilter_accel(double   accel_x,   /* Angular speeds given by IMU */
			    double   accel_y,
			    double   accel_z,
			    double   accelFiltered[IMU_NUM_ACCEL] ) /* output */

{
  int     i, j;
  /* i = number of the previous sample used to filter */  
  Accel samplesSum;


  /* Read the new sample values */
  buffer_accel[k_accel][IMU_ACCEL_X] = accel_x;
  buffer_accel[k_accel][IMU_ACCEL_Y] = accel_y;
  buffer_accel[k_accel][IMU_ACCEL_Z] = accel_z;



  /* *************** FILTER ************************************************/

  /* Weights are applied to the past samples (widthFilter before)
   * Then the current sample is replaced by the weighted average */
  
  /* init samplesSum array */
  for (j = 0; j < IMU_NUM_ACCEL; j++)
    samplesSum[j] = 0.0;

  /* update the sum of the weights */
  if ( K_accel == 0 )  /* first sample */
    /* initial weightsSumF */
    weightsSumF_accel = weightsFilter_accel[0];
  else if (K_accel <= widthFilter_accel)
    /* the current sample is number K < widthFilter ********/
    weightsSumF_accel += weightsFilter_accel[K_accel];

  /* compute the weighted sum of samples */
  if (K_accel <= widthFilter_accel)      /* k = K here */
  {
    for (i = 0; i <= K_accel; i++)   /* i: before the current sample K to modify */
    {
      /* weighted sums */
      for (j = 0; j < IMU_NUM_ACCEL; j++)
	samplesSum[j] = samplesSum[j] + weightsFilter_accel[i]*buffer_accel[k_accel-i][j];
    }
  }
  /* At this moment weightsSumF contains all the weights */
  else  /* the current sample is number K >= widthFilter  ***/
  { 
    /* all the weights can be applied every time */
    /* Thus, weightsSumF is always the same here */

    for (i = 0; i <= k_accel; i++)   /* from k to beginning of buffer */
    {
      /* weighted sum */
      for (j = 0; j < IMU_NUM_ACCEL; j++)
	samplesSum[j] = samplesSum[j] + weightsFilter_accel[i]*buffer_accel[k_accel-i][j];
    }
    for (i = k_accel+1; i <= widthFilter_accel; i++)   /* from end of buffer to k+1 */
    {
      for (j = 0; j < IMU_NUM_ACCEL; j++)
	samplesSum[j] = samplesSum[j] + 
	  weightsFilter_accel[i]*buffer_accel[widthFilter_accel+1+k_accel-i][j];
    }
  }

  /* the sample k is replaced by the weighted average */
  for (j = 0; j < IMU_NUM_ACCEL; j++)
    buffer_accel[k_accel][j] = samplesSum[j] / weightsSumF_accel;

  /* *************** END OF FILTER ****************************************/

  
  /* Get the Filtered values -> Output */
  accelFiltered[IMU_ACCEL_X] = buffer_accel[k_accel][IMU_ACCEL_X];
  accelFiltered[IMU_ACCEL_Y] = buffer_accel[k_accel][IMU_ACCEL_Y];
  accelFiltered[IMU_ACCEL_Z] = buffer_accel[k_accel][IMU_ACCEL_Z];


  
  /* Change k: the number of the new sample in the buffer */
  k_accel = (k_accel+1)%(widthFilter_accel+1);
  K_accel++;


  return OK;
}
