/*	$Id: imuReadTaskCodels.c,v 1.3 2007/04/17 14:38:52 tpeynot Exp $ */

/* 
 * Copyright (c) 2002 LAAS/CNRS                       --  Wed Jun 12 2002
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice,  this list of  conditions and the following disclaimer in
 *      the  documentation  and/or  other   materials provided  with  the
 *      distribution.
 *
 * THIS  SOFTWARE IS PROVIDED BY  THE  COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND  ANY  EXPRESS OR IMPLIED  WARRANTIES,  INCLUDING,  BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES  OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR  PURPOSE ARE DISCLAIMED. IN  NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR      CONTRIBUTORS  BE LIABLE FOR   ANY    DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF  SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE   OF THIS SOFTWARE, EVEN   IF ADVISED OF   THE POSSIBILITY OF SUCH
 * DAMAGE.
 */

#ifndef UNIX
#include <vxWorks.h>
#else
#include <portLib.h>
#endif

#include <time.h>

#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <math.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <netinet/in.h>
#include <fcntl.h>
#include <termios.h>
#include <errno.h>

#include "h2timerLib.h"
#include "commonStructLib.h"

#include "server/imuHeader.h"
#include "imuInt.h"

static double nTicksPerSec;


/*---- Var globales ------*/
/* To know when it is the first cycle (useful for Integration) */
static int imu1stCycleFlag; /* == 1 when First cycle **/
/****Th: to get on line the actual sample time *****/
static double previousTimeSec;




/*---- Useful macros ------*/
#ifndef ABS
#define ABS(x) ((x)<0?-(x):(x))
#endif


/* --- Local functions ----------------------------------------------- */

static STATUS	getData(ImuData *imu, int pktSize);
static STATUS	decodeScaled(ImuData *imu, unsigned char *pkt);
static STATUS	setZero(ImuData *imu);
static STATUS	setMode(ImuData *imu);
static STATUS	setBaud(ImuData *imu, int baud);
static STATUS	expect(ImuData *imu, char e);



/*
 * --- imuReadTaskPerm --------------------------------------------------
 *
 * Communication with the DMU.
 */

STATUS
imuReadTaskPerm(ImuData *imuData)
{
   H2TIMER_ID timer;

   commonStructTake(imuData);

   nTicksPerSec = sysClkRateGet();

   /* create and start timer */
   if ((timer = h2timerAlloc()) == NULL) {
      fputs("imuReadTaskInit: h2timerAlloc", stderr);
      commonStructGive(imuData);
      return OK;
   }

   if (h2timerStart(timer, imuData->period, 0) != OK) {
      fputs("imuReadTaskInit: h2timerStart", stderr);
      commonStructGive(imuData);
      return OK;
   }

   /* configure baud rate */
   if (imuData->targetBaud != imuData->baud) {
#if 0
      if (setBaud(imuData, imuData->targetBaud) != OK) {
	 commonStructGive(imuData);
	 return OK;
      }
#endif
      imuData->baud = imuData->targetBaud;
   }

   if (setMode(imuData) != OK) {
      commonStructGive(imuData);
      return OK;
   }

   imuData->exit = 0;
   commonStructGive(imuData);

   /* Process commands */
   while(!imuData->exit) {
      if (h2timerPause(timer) != OK) {
	 fputs("imuReadTask: h2timerPause failed\n", stderr);
	 break;
      }

      commonStructTake(imuData);

      if (getData(imuData, 18) != OK) {
/* 	 commonStructGive(imuData); */
	 fputs("imuReadTask: getData failed\n", stderr);
	 continue;
      }

      if (imuData->zero > 0.0) {
	 if (setZero(imuData) != OK) {
	    imuData->zero = -1.0;
	 } else {
	    imuData->zero = 0.0;
	 }
      }

      commonStructGive(imuData);
   }

   commonStructTake(imuData);
   h2timerFree(timer);
   imuData->exit = 0;
   commonStructGive(imuData);
   return OK;
}




/*
 * --- integrateData ----------------------------------------------------
 *
 */

static STATUS
integrateData(ImuData *imu)
{

  double actualSampleTime;
  double xAccel_LamaFrame, yAccel_LamaFrame, zAccel_LamaFrame;



  /* Calculate the actual sample time: previousTime-currentTime */
  actualSampleTime = imu->time - previousTimeSec;

  /* Original integration, by Anthony */
/*    imu->roll += imu->rollRate * imu->period/nTicksPerSec; */
/*    imu->pitch += imu->pitchRate * imu->period/nTicksPerSec; */
/*    imu->yaw += imu->yawRate * imu->period/nTicksPerSec; */

   /* Should use the actual sampleTime: previousTime-currentTime */
   imu->roll += imu->rollRateNoBias * actualSampleTime;
   imu->pitch += imu->pitchRateNoBias * actualSampleTime;
   imu->yaw += imu->yawRateNoBias * actualSampleTime;

   /* Should be The Same for Accelerations... */
   imu->xSpeed += imu->xAccelNoBias * imu->period/nTicksPerSec;
   imu->ySpeed += imu->yAccelNoBias * imu->period/nTicksPerSec;
   imu->zSpeed += imu->zAccelNoBias * imu->period/nTicksPerSec;

   /***** Variances *****/
   /* Var(int.X_k) = Var(int.X_k-1) + dt^2 * Var(Xk) */
   imu->xSpeedVariance += SDI_F->imuVariances.xAccelVariance * SQ(imu->period/nTicksPerSec);
   imu->ySpeedVariance += SDI_F->imuVariances.yAccelVariance * SQ(imu->period/nTicksPerSec);
   imu->zSpeedVariance += SDI_F->imuVariances.zAccelVariance * SQ(imu->period/nTicksPerSec);
   /* Roll, pitch & yaw */
   imu->rollVariance += SDI_F->imuVariances.rollRateVariance * SQ(imu->period/nTicksPerSec);
   imu->pitchVariance += SDI_F->imuVariances.pitchRateVariance * SQ(imu->period/nTicksPerSec);
   imu->yawVariance += SDI_F->imuVariances.yawRateVariance * SQ(imu->period/nTicksPerSec);
   
   /* Position in x, y ,z */
   imu->xPos += imu->xSpeed * imu->period/nTicksPerSec;
   imu->yPos += imu->ySpeed * imu->period/nTicksPerSec;
   imu->zPos += imu->zSpeed * imu->period/nTicksPerSec;
   /* Variances for the Position */
   /** A verifier: ne vaut-il pas mieux faire ça à partir des accelVariances? **/
   imu->xPosVariance += imu->xSpeedVariance * SQ(imu->period/nTicksPerSec);
   imu->yPosVariance += imu->ySpeedVariance * SQ(imu->period/nTicksPerSec);
   imu->zPosVariance += imu->zSpeedVariance * SQ(imu->period/nTicksPerSec);
   /***** Th End ************************************/


   /*** Calculation of the attitude angles using only the accelerations measured (OK when Dala does not move) ********/
   /** First: transformation to get the accelerations in the usual frame on LAMA **/
   xAccel_LamaFrame = - imu->yAccelFiltered;
   yAccel_LamaFrame = imu->xAccelFiltered;
   zAccel_LamaFrame = imu->zAccelFiltered;
   /** These accelerations are used to calculate the roll and pitch: **/
   /* accel should all be in [-1;1] */
   if( ABS(xAccel_LamaFrame) > 1 )
     {
       printf("imuAttitudeAngles: xAccel is over limit\n");
       imu->phiLama = 0.0;
     }
   else
     imu->phiLama = asin(xAccel_LamaFrame);
   /*----------------*/
   if( ABS(yAccel_LamaFrame) > 1 )
     {
       printf("imuAttitudeAngles: yAccel is over limit\n");
       imu->psiLama = 0.0;
     }
   else
     imu->psiLama = asin( -yAccel_LamaFrame/cos(imu->phiLama) );
   /** Check that relation between az and phi, psi works ?? */


   return OK;
}



/*
 * --- filter for imuData ----------------------------------------------------
 *
 */

static STATUS
filterData(ImuData *imu)
{
  double omegaFiltered[IMU_NUM_ANGSPEEDS];
  double accelFiltered[IMU_NUM_ACCEL];


  /* Angular Speeds Filter activated */
  /* The interesting data to filter are the ang. speeds without bias */
  if(SDI_F->imuDataFilterParams.angularSpeedsFilterMode == IMU_FILTER_ON)
    {
      if( imuGaussAverageFilter(imu->rollRateNoBias, imu->pitchRateNoBias, 
				imu->yawRateNoBias,  omegaFiltered ) != OK )
	return ERROR;
      
      /* Copy of results in imuData */
      imu->rollRateFiltered = omegaFiltered[IMU_OMEGA_X];
      imu->pitchRateFiltered = omegaFiltered[IMU_OMEGA_Y];
      imu->yawRateFiltered = omegaFiltered[IMU_OMEGA_Z];
    }
  /* Filter Not Activated: copy the data without bias, but not filtered*/
  else
    {
      imu->rollRateFiltered  = imu->rollRateNoBias;
      imu->pitchRateFiltered = imu->pitchRateNoBias;
      imu->yawRateFiltered   = imu->yawRateNoBias;
    }

  /*-----------------------------------------*/

  /* Accelerations Filter activated */
  /* The interesting data to filter are the acceleration without bias */
  if(SDI_F->imuDataFilterParams.accelFilterMode == IMU_FILTER_ON)
    {
      if( imuGaussAverageFilter_accel(imu->xAccelNoBias, imu->yAccelNoBias, 
			    imu->zAccelNoBias, accelFiltered ) != OK )
	return ERROR;
      
      /* Copy of results in imuData (A creer) */
      imu->xAccelFiltered = accelFiltered[IMU_ACCEL_X];
      imu->yAccelFiltered = accelFiltered[IMU_ACCEL_Y];
      imu->zAccelFiltered = accelFiltered[IMU_ACCEL_Z];
    }
  /* Filter Not Activated: copy the data without bias, but not filtered */
  else
    {
      imu->xAccelFiltered = imu->xAccelNoBias;
      imu->yAccelFiltered = imu->yAccelNoBias;
      imu->zAccelFiltered = imu->zAccelNoBias;
    }


   return OK;
}




/*
 * --- getData ----------------------------------------------------------
 *
 * Get DMU data: send 'G', receive packet
 */

static STATUS
getData(ImuData *imu, int pktSize)
{
   int max = 20;
   int size, n;
   unsigned char pkt[32], *p, chk;
   char command;

   command = 'G';
   if (write(imu->fd, &command, 1) != 1) {
      perror("getData");
      return ERROR;
   }

   size = read(imu->fd, pkt, pktSize);
   if (size <= 0) {
      perror("getData");
      return ERROR;
   }

   do {
      /* receive a complete packet */
      while(*pkt != 0xff && size > 0) {
	 size--;
	 memmove(pkt, pkt+1, size);
      }

      if (size != pktSize) do {
	 n = read(imu->fd, pkt+size, pktSize-size);
	 if (imu->exit) {
	   printf ("imu read : exit\n");
	   /* close fd here ? */
	   /* return OK or ERROR ? */
	   return ERROR;
	 }
	 if (n == 0) {
	    if (write(imu->fd, &command, 1) != 1) {
	       perror("getData");
	       return ERROR;
	    }
	 }
	 if (n < 0) {
	    perror("getData");
	    return ERROR;
	 }
	 size += n;
      } while(size != pktSize);

      /* checksum */
      for(n=1,chk=0, p=pkt+1; n<size-1; n++, p++) chk+=*p;
      if (chk == pkt[size-1]) return decodeScaled(imu, pkt);

      *pkt = 0;
   } while(max--);

   return ERROR;
}


/*
 * --- decodeScaled -----------------------------------------------------
 *
 * Decode a packet in scaled format
 */

static STATUS
decodeScaled(ImuData *imu, unsigned char *pkt)
{
  /* Struct for time */
  struct timespec tp;
  u_long msec;
  double sec;
#ifdef ATRV_JUNIOR
  double xAccelGromit, yAccelGromit, zAccelGromit;
  double wx_gromit, wy_gromit, wz_gromit;
  /*****************/

/*    commonStructTake(imu); */
  
#endif

   /* decode packet */
   /* first w is omega_x = pitchRate (check sign/Lama) */
   imu->pitchRate = (signed short)ntohs(*(short *)(&pkt[1]));
   /* second w is omega_y = rollRate (check sign/Lama) */
   imu->rollRate = (signed short)ntohs(*(short *)(&pkt[3]));
   /* third w is omega_z = yawRate */
   imu->yawRate = (signed short)ntohs(*(short *)(&pkt[5]));

   imu->xAccel = (signed short)ntohs(*(short *)(&pkt[7]));
   imu->yAccel = (signed short)ntohs(*(short *)(&pkt[9]));
   imu->zAccel = (signed short)ntohs(*(short *)(&pkt[11]));

   imu->temp = ntohs(*(short *)(&pkt[13]));
   imu->ticks = ntohs(*(short *)(&pkt[15]));

#ifdef ATRV_JUNIOR
   /* Conversion needed between Gromit's and Dala's IMU axis */
   /* first, copy all */
   xAccelGromit = imu->xAccel;
   yAccelGromit = imu->yAccel;
   zAccelGromit = imu->zAccel;
   wx_gromit    = imu->pitchRate;
   wy_gromit    = imu->rollRate;
   wz_gromit    = imu->yawRate;

   /* then, make the conversion */
   imu->xAccel = yAccelGromit;
   imu->yAccel = xAccelGromit;
   imu->zAccel = -zAccelGromit;

   imu->pitchRate = wy_gromit;
   imu->rollRate  = wx_gromit;
   imu->yawRate   = -wz_gromit;
#endif

   /*--- scale (IMU400CC-200) -------------------*/
   /* Rate(deg/s) = data * RR * 1.5/2^15  */
   /* => Rates in rad/s :  */
   imu->rollRate *= (ANGULAR_RATE_RANGE * 1.5)/(1<<15) * M_PI/180;
   imu->pitchRate *= (ANGULAR_RATE_RANGE * 1.5)/(1<<15) * M_PI/180;
   imu->yawRate *= (ANGULAR_RATE_RANGE * 1.5)/(1<<15) * M_PI/180;
   /* Accel(G) = data * GR * 1.5/2^15  */
   imu->xAccel *= (G_RANGE * 1.5)/(1<<15);
   imu->yAccel *= (G_RANGE * 1.5)/(1<<15);
   imu->zAccel *= (G_RANGE * 1.5)/(1<<15);

   /* Temperature in deg C */
   imu->temp = 44.4 * (imu->temp * 5.0/4096 - 1.375);

   /*** Time ***/
   clock_gettime(CLOCK_REALTIME, &tp);
   msec = (tp.tv_nsec / 1000000) + (tp.tv_sec * 1000);
   sec  = ((double)(msec))/1000;
   
   imu->time = sec;
   /******************/

   /** The first time one enters this function, integration cannot be done -> just record current time value for next time **/
   /** After the first time, integrate imuData */
   if( imu1stCycleFlag == 0 )
     integrateData(imu);
   else
     imu1stCycleFlag = 0;
   
   /* Record the current time value as previous time for next time */
   previousTimeSec = sec;
     

   /** Bias compensation for the accelerations **/
   /** Default values for the biases: 0.0 **/
   imu->xAccelNoBias = imu->xAccel - SDI_F->sensors.xAccelBias;
   imu->yAccelNoBias = imu->yAccel - SDI_F->sensors.yAccelBias;
   imu->zAccelNoBias = imu->zAccel - SDI_F->sensors.zAccelBias;

   /** Bias compensation for the angular speeds **/
   /** Default values for the biases: 0.0 **/
   imu->rollRateNoBias  = imu->rollRate  - SDI_F->sensors.rollRateBias;
   imu->pitchRateNoBias = imu->pitchRate - SDI_F->sensors.pitchRateBias;
   imu->yawRateNoBias   = imu->yawRate   - SDI_F->sensors.yawRateBias;

   /** Filter of Ang Speeds and/or accelerations (if activated only) ***/
   /** Default: copy raw data */
   filterData(imu);
   /*******************/


/*    commonStructGive(imu); */
   return OK;
}


/*
 * --- setZero ----------------------------------------------------------
 *
 * Compute bias
 */

static STATUS
setZero(ImuData *imu)
{
   STATUS s;
   struct termios t;
   unsigned char command[2];

   command[0] = 'z';
   command[1] = imu->zero/7 * 100; /* about 7ms per sample */
   if (write(imu->fd, command, 2) != 2) {
      perror("setMode");
      return ERROR;
   }

   fprintf(stderr, "imu: zeroing for %d\n", command[1]);

   /* configure read in blocking mode (zeroing might be long) */
   tcgetattr(imu->fd, &t);
   t.c_cc[VMIN] = 1;
   t.c_cc[VTIME] = 0; /* blocking read */
   tcsetattr(imu->fd, TCSANOW, &t);

   s = expect(imu, 'Z'); /* two Z */
   s = expect(imu, 'Z');

   t.c_cc[VMIN] = 0;
   t.c_cc[VTIME] = 1; /* blocking read, 100ms timeout */
   tcsetattr(imu->fd, TCSANOW, &t);

   if (s != OK) {
      fputs("zero failed\n", stderr);
      return ERROR;
   }

   s = getData(imu, 18);
   if (s != OK) {
      fputs("getData failed for static angles\n", stderr);
      return ERROR;
   }

   imu->yaw = 0.0;
   imu->roll = atan2(-imu->yAccel, -imu->zAccel);
   imu->pitch = atan2(imu->xAccel, -imu->zAccel);

   /* Init time ----------------------------- */
   imu->time += 0;
   /* --------------------------------------- */

   return OK;
}


/*
 * --- setMode ----------------------------------------------------------
 *
 * For now: set scaled mode.
 */

static STATUS
setMode(ImuData *imu)
{
   char command;

   /** Command to have Data in Scaled Mode **/
   command = 'c';
   if (write(imu->fd, &command, 1) != 1) {
      perror("setMode");
      return ERROR;
   }
   /** Expect Normal Answer to that command: 'C' **/
   if (expect(imu, 'C') != OK) {
      fputs("setMode failed\n", stderr);
      return ERROR;
   }

   return OK;
}


/*
 * --- setBaud ----------------------------------------------------------
 *
 * Change baud rate:
 * send 'b', configure serial port, send 'a', receive 'D'. 
 */

static STATUS
setBaud(ImuData *imu, int baud)
{
   char command;
   struct termios t;

   /* send 'b' */
   command = 'b';
   if (write(imu->fd, &command, 1) != 1) {
      perror("setBaud");
      return ERROR;
   }

   /* configure serial port */
   tcgetattr(imu->fd, &t);
   cfsetospeed(&t, baud);
   cfsetispeed(&t, baud);
   tcsetattr(imu->fd, TCSANOW, &t);

   /* send 'a' */
   command = 'a';
   if (write(imu->fd, &command, 1) != 1) {
      perror("setBaud");
      return ERROR;
   }

   /* expect 'D' */
   if (expect(imu, 'D') != OK) {
      fputs("setBaud failed\n", stderr);
      return ERROR;
   }

   return OK;
}


/*
 * --- expect -----------------------------------------------------------
 *
 * Wait for a character
 */

static STATUS
expect(ImuData *imu, char e)
{
   int n, max = 20;
   char c;

   while(max--) {
      n = read(imu->fd, &c, 1);
      if (n == 1 && c == e) return OK;
      if (n < 0 && errno != EINTR) break;
   }

   return ERROR;
}
 


/************************************************************/
/**** Initialisations to be done when module starts ******/
/*    (function called by codel imuExecTaskInit)     */
/************************************************************/

STATUS
imuInitGlobalVariables()
{
  
  imu1stCycleFlag = 1;
    
  return OK;
}



/*
 * --- get Version ----------------------------------------------------------
 *
 * Get DMU Version: send 'v', receive ASCII String: DMU type and firmware version
 */

STATUS
getVersion(ImuData *imu)
{
  int size;
  char pkt[DEVICE_NAME_LEN];
  int  pktSize=DEVICE_NAME_LEN;
  char command;
  unsigned int serialnb;

  command = 'v';
   if (write(imu->fd, &command, 1) != 1) {
      perror("getVersion");
      return ERROR;
   }

   size = read(imu->fd, pkt, pktSize);
   if (size <= 0) {
     printf("imu: getVersion: size read is: %d\n", size);
      perror("getVersion");
      return ERROR;
   }
   if (size > pktSize) {
     printf("imu: getVersion: size > pktSize\n");
      perror("getVersion");
      return ERROR;
   }
  
   printf("imu: getVersion: received pkt size is: %d\n", size );

   /*-- Copy result obtained in struct imuData --*/
   strncpy(imu->version, pkt, DEVICE_NAME_LEN);


   /*-- Read serial number ---*/
   command = 'S';
   if (write(imu->fd, &command, 1) != 1) {
      perror("getSerialNb");
      return ERROR;
   }
   size = read(imu->fd, &serialnb, 4);
   if (size <= 0) {
     printf("imu: getSerialNb: size read is: %d\n", size);
      perror("getSerialNb");
      return ERROR;
   }

/*    imu->version = (long) serialnb; */

   printf("imu: getSerial: received pkt size is: %d\n", size );
   printf("imu: Serial Nb: %d\n", serialnb);

   return OK;
}
