/**************************************************************
 **
 ** Copyright (c) 2004-2005 LAAS/CNRS 
 **
 ** Thierry Peynot - 2004-2005
 **
 ** $Source: /cvs/openrobots/modules/imu/lib/backstepping.c,v $
 ** $Revision: 1.1 $
 ** $Date: 2007/04/16 14:11:06 $
 ** 
 **************************************************************/


/* standard librairies */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "libimu.h"


/* useful macros */
#include <math.h>


/* Dimension des vecteurs */
#define IMU_VECT_DIM 3


/*********** Declaration of functions ***************/
int
imuScalarMultVect(double  vect[IMU_VECT_DIM],
		  double  scalar );

int
imuScalarMultNewVect(double  vectMult[IMU_VECT_DIM],
		     double  vect[IMU_VECT_DIM],
		     double  scalar );


int
imuVectDiffNew(double  vectDiff[IMU_VECT_DIM],
	       double  vect1[IMU_VECT_DIM],   
	       double  vect2[IMU_VECT_DIM]  );

int
imuVectPlusNew(double  vectPlus[IMU_VECT_DIM],
	       double  vect1[IMU_VECT_DIM],  
	       double  vect2[IMU_VECT_DIM]  );



int
imuCopyVect(double  vectCopied[IMU_VECT_DIM],
	    double  vect[IMU_VECT_DIM] );

int
imuCrossProduct(double  vectC[3],
		double  vectA[3],
		double  vectB[3] );

int
imuNormalizeVector(double  vect[3]);

int
imuNewNormalizedVector(double  normVect[3],
		       double  vect[3] );

int
imuEulerIntegrate(double  intVect[3], 
		  double  vect[3],    
		  double  timePerSample);




/**************************************************************************************
 *
 * Multiplication of an input vector[IMU_VECT_DIM] by a scalar
 * The input vector IS NOT MODIFIED : The output vector is: scalar * vect
 *
 **************************************************************************************/

int
imuScalarMultNewVect(double  vectMult[IMU_VECT_DIM],  /* output vector */
		     double  vect[IMU_VECT_DIM],      /* input vector */
		     double  scalar )
{
  int  i;
  
  for(i=0; i<IMU_VECT_DIM; i++)
    vectMult[i] = scalar * vect[i];

  return IMU_OK;
}







/**************************************************************************************
 *
 * Multiplication of an input vector[IMU_VECT_DIM] by a scalar
 * The input vector IS MODIFIED : scalar * vect
 *
 **************************************************************************************/

int
imuScalarMultVect(double  vect[IMU_VECT_DIM],   /* input / output vector */
		  double  scalar )
{
  int  i;
  
  for(i=0; i<IMU_VECT_DIM; i++)
    vect[i] = scalar * vect[i];

  return IMU_OK;
}



/**************************************************************************************
 *
 * Difference between 2 input vectors[IMU_VECT_DIM]
 * Output vector : vectDiff = vect1 - vect2
 *
 **************************************************************************************/

int
imuVectDiffNew(double  vectDiff[IMU_VECT_DIM],  /* output vector */
	       double  vect1[IMU_VECT_DIM],     /* input vector 1 */
	       double  vect2[IMU_VECT_DIM]  )     /* input vector 2 */
{
  int  i;
  
  for(i=0; i<IMU_VECT_DIM; i++)
    vectDiff[i] = vect1[i] - vect2[i];

  return IMU_OK;
}



/**************************************************************************************
 *
 * Addition of 2 input vectors[IMU_VECT_DIM]
 * Output vector : vectPlus = vect1 + vect2
 *
 **************************************************************************************/

int
imuVectPlusNew(double  vectPlus[IMU_VECT_DIM],  /* output vector */
	       double  vect1[IMU_VECT_DIM],     /* input vector 1 */
	       double  vect2[IMU_VECT_DIM]  )     /* input vector 2 */
{
  int  i;
  
  for(i=0; i<IMU_VECT_DIM; i++)
    vectPlus[i] = vect1[i] + vect2[i];

  return IMU_OK;
}







/**************************************************************************************
 *
 * Make a copy of an input vector[IMU_VECT_DIM]
 *
 **************************************************************************************/

int
imuCopyVect(double  vectCopied[IMU_VECT_DIM],   /* output vector: copy */
	    double  vect[IMU_VECT_DIM] )        /* input vector to copy */
{
  int  i;
  
  for(i=0; i<IMU_VECT_DIM; i++)
    vectCopied[i] = vect[i];

  return IMU_OK;
}







/**************************************************************************************
 *
 * Compute the cross product of 2 vectors (3 dimensions)
 * vectC (output) = vectA X vectB
 *
 **************************************************************************************/

int
imuCrossProduct(double  vectC[3],   /* output */
		double  vectA[3],
		double  vectB[3] )
{

  vectC[0] = vectA[1]*vectB[2] - vectA[2]*vectB[1];
  vectC[1] = vectA[2]*vectB[0] - vectA[0]*vectB[2];
  vectC[2] = vectA[0]*vectB[1] - vectA[1]*vectB[0];

  return IMU_OK;
}





/**************************************************************************************
 *
 * Normalization of the input vector
 * 
 * This input vector is MODIFIED to become a Norme 1 vector (using Euclidian norm)
 *
 **************************************************************************************/

int
imuNormalizeVector(double  vect[3])    /* input / output */
{
  double  norme;
  int     i;

  norme = sqrt( SQ(vect[0]) + SQ(vect[1]) + SQ(vect[2]) );
  
  if( norme == 0.0 )
    {
      for(i=0; i<3; i++)
	vect[i] = 0.0;
    }
  else
    {
      for(i=0; i<3; i++)
	vect[i] = vect[i] / norme;
    }

  return IMU_OK;
}



/**************************************************************************************
 *
 * Normalization of an input vector
 * 
 * This input vector is NOT MODIFIED
 * The corresponding Norm 1 vector is given as an output
 *
 **************************************************************************************/

int
imuNewNormalizedVector(double  normVect[3],  /* output */
		       double  vect[3])    /* input */
{
  double  norme;
  int     i;

  norme = sqrt( SQ(vect[0]) + SQ(vect[1]) + SQ(vect[2]) );
  
  if( norme == 0.0 )
    {
      for(i=0; i<3; i++)
	normVect[i] = 0.0;
    }
  else
    {
      for(i=0; i<3; i++)
	normVect[i] = vect[i] / norme;
    }

  return IMU_OK;
}





/**************************************************************************************
 *
 * Standard (and silly) Euler Discrete Integration of a 3 dimension vector
 *  (dim could be N)
 *
 * 'intVect' is the sum of 'vect' over the time 
 *
 * ***Beware!!**** the vector intVect must have been initialized before!!!
 *
 **************************************************************************************/

int
imuEulerIntegrate(double  intVect[IMU_VECT_DIM],   /* input / output: result of integration */
		  double  vect[IMU_VECT_DIM],      /* Input: Data to integrate */
		  double  timePerSample)  /* Sampling period */
{
  
  int  i;

  for(i=0; i<IMU_VECT_DIM; i++)
    intVect[i] += vect[i] * timePerSample;

  
  return IMU_OK;
}




/**************************************************************************************
 *
 * Conversion d'un vecteur de RepereIMUdala -> Repere aero (utilise pour estime la gavite)
 * 
 * Vecteur en INPUT est MODIFIE
 *
 **************************************************************************************/

int
imuDalaImu2aeroFrame(double  vect[3] )   /* input / output: result of conversion */
{
  
  double  temp[3];

  imuCopyVect(temp, vect);
  
  vect[0] = - temp[1];
  vect[1] = - temp[0];
  vect[2] = - temp[2];
  
  return IMU_OK;
}






/**************************************************************************************
 *
 * Compute the 3 attitude angles, given an estimation of the gravity vector (Xest)
 *  Xest will be named 'gravity'
 *
 * angles[0] = pitch angle = asin(gravity_x)
 * angles[1] = roll angle  = -asin( gravity_y / cosPitch )
 * angles[2] = roll angle computed differently  = -acos( gravity_z / cosPitch )
 *
 **************************************************************************************/

int
imuAttitudeAngles(double  angles[2],   /* output: the attitude angles */
		  double  gravityVect[3] )  /* Estimation of the gravity Vector */
{
  int imuErr;
  double pitch;
  double cosPitch;

  pitch = asin( gravityVect[0] );
  cosPitch = cos(pitch);

  angles[0] = pitch;

/*   if( cosPitch == 0.0 ) */
/*     { */
/*       imuErr = error (ERR_IMU_OFFLIMITS); */
/*       printf("califeIMU: cos(pitch)=0.0: Pitch is over limit !!\n" ); */
/*     } */
/*   else */
/*     { */
/*       angles[1] = asin( gravityVect[1] / cosPitch ); */
/*       angles[2] = -acos( gravityVect[2] / cosPitch ); */
/*       imuErr = IMU_OK; */
/*     } */

  if(gravityVect[2] != 0.0)
    {
      angles[1] = atan2(gravityVect[1], gravityVect[2]);
      imuErr = IMU_OK;
    }
  else
    {
      imuErr = error (ERR_IMU_OFFLIMITS);
      printf("califeIMU: gravityVect[2]=0.0: Pitch is over limit !!\n" );
    }
  
  return imuErr;
}





/**************************************************************************************
 *
 * Computation of an estimation of the gravity vector at time sample k
 * given (Input): 
 * - mes_gyro     = the 3 angular speeds provided by the IMU gyros
 * - mes_accel = the 3 acceleration provided by the IMU accelerometers
 * - gravityPrev  = the gravity vector estimated at previous time sample: (k-1)
 *
 * Other Inputs:
 * - gain Km
 * - gain Kb (it seems that generally Km = 10*Kb)
 *
 * Output:
 * - gravity:  the estimated gravity vector at k
 *   !!!!!! gravity doit etre initialise avant le 1er appel a cette fonction!!!
 * - Bestimated: the 3 estimated biases for the gyros
 *   !!!!!! Must be initialized too!!
 **************************************************************************************/


int
imuGravityEstimation(double  gravity[3],   /* OUTPUT: the estimated gravity vector */
		     /* Must be initialized !! */
		     double  Bestimated[3],   /* OUTPUT: the 3 gyros estimated biases */
		     /* Must be initialized !! */
		     double  mes_gyro[3],  /* the 3 omega measured by the IMU gyros */
		     double  mes_accel[3],  /* the 3 accel measured by the IMU acceleros */
		     double  Km,     /* Explications ? */
		     double  Kb,     /* Explications ? */
		     double  timePerSample )  /* Sampling Period */
{

  double  gravityPrev[3];  /* The previous gravity vector estimated */
  double  crossGravity[3];
  double  temp[3];
  double  temp1[3];
  double omega[3];
  

  /*** Conversions de repere faites POUR LE MOMENT en interne ****/
  imuDalaImu2aeroFrame(mes_gyro);
  imuDalaImu2aeroFrame(mes_accel);
  /***************************************************************/


  /* First, gravityPrev is gravity entered */
  imuCopyVect(gravityPrev, gravity);


  imuNormalizeVector(mes_accel);
  /* crossGravity = Xes x Xmes = gravityPrev x mes_accel */
  imuCrossProduct(crossGravity, gravityPrev, mes_accel);

  /* Calcul de Best = int( Kb*(crossGravity) ) */
  imuScalarMultNewVect(temp, crossGravity, Kb);
  imuEulerIntegrate(Bestimated, temp, timePerSample);

  /* Calcul de omega = Km*(crossGravity) */
  imuScalarMultNewVect(omega, crossGravity, Km);
  
  /* temp = omega + Best */
  imuVectPlusNew(temp, Bestimated, omega);
  /* temp = mes_gyro - (omega + Bestimated) */
  imuVectDiffNew(temp, mes_gyro, temp);

  /* temp1 = Xest x (mes_gyro - (omega + Bestimated)) */
  imuCrossProduct(temp1, gravityPrev, temp);
  /* Integration de temp1 pour obtenir gravity (non norme) */
  imuEulerIntegrate(gravity, temp1, timePerSample);

  /* Result */
  imuNormalizeVector(gravity);
      

  return IMU_OK;
}
