/**************************************************************
 **
 ** Copyright (c) 2006 LAAS/CNRS 
 **
 ** Thierry Peynot - 2006
 **
 ** $Source: /cvs/openrobots/modules/imu/lib/monitor.c,v $
 ** $Revision: 1.2 $
 ** $Date: 2007/06/29 15:13:30 $
 ** 
 **************************************************************/

/* standard librairies */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "libimu.h"


/* useful macros */
#include <math.h>


/* Global variables */
static float pitchRateQuadSum;  /* Quadratic sum */
static float rollRateQuadSum;   /* Quadratic sum */

static float* pitchRateBuffer;  /* contains the necessary signal history */
static float* rollRateBuffer;  /* contains the necessary signal history */

static int K;  /* Sample Number */





/**************************************************************************************
 *
 * Inverse Sigmoid function (for a soft threshold)
 *  (1-sigmoid)
 *
 **************************************************************************************/

float
imu_sigmoid (float    x,
	     float    m,        /* m = threshold */
	     float    sigma )    /* cf. Standard deviation  */
  
{
  return ( 1.0 - 1.0/(1.0+exp(-2.0*(m-x)/sigma)) );
}




/**************************************************************************************
 *
 * Computation of a the current signal energy 
 *
 **************************************************************************************/
int
imuComputeEnergyInit(int  nbCycles)  /* Nb of cycles for the time window */
{

  /* Initialize sample number */
  K = 0;

  pitchRateQuadSum = 0.0;
  rollRateQuadSum = 0.0;

  /*---- allocation of  buffers ----*/
  if (pitchRateBuffer != NULL) free(pitchRateBuffer);
  if (rollRateBuffer  != NULL) free(rollRateBuffer);

  pitchRateBuffer = calloc (nbCycles, sizeof (float));
  if (pitchRateBuffer  == NULL ) {
    return error (ERR_IMU_MALLOC);
  }

  rollRateBuffer = calloc (nbCycles, sizeof (float));
  if (rollRateBuffer  == NULL ) {
    return error (ERR_IMU_MALLOC);
  }
  

  return(IMU_OK);
}




/**************************************************************************************
 *
 * Computation of pitchRate and rollRate energies 
 *
 **************************************************************************************/
int
imuComputeEnergy(float*  pitchRateEnergy,  /* output: new energy computed */
		 float*  rollRateEnergy,
		 float   pitchRate,  /* New value for the current signal */
		 float   rollRate, 
		 int     timeWidth_nbCycles)  /* Nb of Cycles = width of time window */
{
  float  sqPitchRate, sqRollRate;
  int  i;

  if(timeWidth_nbCycles <= 0) {
    fprintf(stderr,"imuComputeEnergy: timeWidth_nbCycles=%d not valid \n", timeWidth_nbCycles);
    return error(ERR_IMU_BADPARAMS);
  } 

  /* new squares needed */
  sqPitchRate = SQ(pitchRate);
  sqRollRate  = SQ(rollRate);
  
  /* For the first sample, need to sum */
  if( K < timeWidth_nbCycles) {
    /* Save balue in buffer */
    pitchRateBuffer[K] = sqPitchRate;
    rollRateBuffer[K]  = sqRollRate;
    /* update the Quadratic Sum needed to compute the energy */
    pitchRateQuadSum += sqPitchRate;
    rollRateQuadSum  += sqRollRate;

    /* Generate (partial) output */
    *pitchRateEnergy = pitchRateQuadSum / ((float) (K+1));
    *rollRateEnergy  = rollRateQuadSum  / ((float) (K+1));
    
    K++;  /* Incrementation of current sample number */

  } /* Now correct Energy computation can be completed */
  else {
    /* Remove the oldest sample from the sums */
    pitchRateQuadSum -= pitchRateBuffer[0];
    rollRateQuadSum  -= rollRateBuffer[0];

    /* Add the new one */
    pitchRateQuadSum += sqPitchRate;
    rollRateQuadSum  += sqRollRate;

    /*------- Buffer operations --------------------*/
    /* One sample shift in  buffers (left) : remove first element and add new one at the end */
    for(i=0; i < timeWidth_nbCycles-1; i++) {
      pitchRateBuffer[i] = pitchRateBuffer[i+1];
      rollRateBuffer[i]  = rollRateBuffer[i+1];
    }
    /* Add the new sample in buffer */
    pitchRateBuffer[timeWidth_nbCycles-1] = sqPitchRate;
    rollRateBuffer[timeWidth_nbCycles-1]  = sqRollRate;
    /*----------------------------------------------*/
    
    /* Generate output */
    *pitchRateEnergy = pitchRateQuadSum / ((float) timeWidth_nbCycles);
    *rollRateEnergy  = rollRateQuadSum  / ((float) timeWidth_nbCycles);
  }
  

  return IMU_OK;
}
